/*
 * $Id: ring.c,v 1.4 1997/05/08 18:45:14 veran Exp $
 */

/*
 * ring.c
 *
 * data structure methods for a two-d manifold
 * effectively, concentric rings
 *
 */
#include <stdlib.h>
#include <stdio.h>

#include "memory.h"
#include "math_utils.h"
#include "plot.h"
#include "ring_local.h"

/*
 * create a new single ring with n elements in it
 */
ring
ring_create(int n, int dim)
{
  int i;
  ring last, next;

  ring r = (ring) calloc(n, sizeof(RING_ITEM));
  double *x = dvector(0,n*dim-1);

  last = NULL;
  for (i=0; i<n; i++)
    {
      if (i< n-1) next = r+i+1;
      else next = r;
      r[i].index = i;
      r[i].in = NULL;
      r[i].out = NULL;
      r[i].right = next;
      r[i].left = last;
      r[i].x_dim = dim;
      r[i].x = x+i*dim;
      r[i].size = n;
      last = r+i;
    }
  r[0].left = last;

  return r;
}

/*
 * frees up a whole ring!
 */
void
ring_free(ring r)
{
  ring outer, current, temp;

  /* start from the inner ring and work out */
  r = ring_inner(r);
  while (r)
    {
      outer = r->out;
      current = r;
      /* remove the inserted elements */
      do {
	if (current->index < 0)
	  {
	    free_dvector(current->x, 0, current->x_dim);
	    temp = current->right;
	    free(current);
	    current = temp;
	  }
      } while (current != r);
      /* free the rest of the memory */
      free_dvector(r->x,0,r->size * r->x_dim - 1);
      free(r);
      r = outer;
    }
}

/*
 * frees ring and any interior rings!
 *
 *   this proc cleanly detaches the ring from any outer rings
 */
void
ring_free_interior(ring r)
{
  ring current;

  /* check for null ring */
  if (!r) return;

  /* detach from outer ring */
  current = r;
  do {
    if (current->out) 
      {
	current->out->in = NULL;
	current->out = NULL;
      }
    current = current->right;
  } while (current != r);

  ring_free(r);
}

/*
 * excises a ring cleanly
 */
void
ring_excise(ring r)
{
  ring current;

  /* check for null ring */
  if (!r) return;

  /* detach from outer ring */
  current = r;
  do {
    if (current->out)
      {
	current->out->in = current->in;
      }
    if (current->in)
      {
	current->in->out = current->out;
      }
    current->out = NULL;
    current->in = NULL;
    current = current->right;
  } while (current != r);

  ring_free(r);
}

void
ring_extend(ring r)
{
  ring new_ring, base = NULL;
  int i, n;

  /* first find the outer ring and how many items there are */
  base = r;  /* find the base */
  while (base->index != 0) base=base->right;
  while (base->out != NULL) base=base->out;
  if (base->index != 0)
    fprintf(stderr,"extend_ring - ring structure corrupted!");
  new_ring = base;
  n =0;
  do {
    n++;
    new_ring = new_ring->right;
  } while (new_ring != base);

  /* get new ring and attach them */
  new_ring = ring_create(n, base->x_dim);
  for (i=0; i<n; i++)
    {
      new_ring->in = base;
      base->out = new_ring;
      new_ring = new_ring->right;
      base = base->right;
    }
}

void
ring_dump(ring r)
{
  ring r2;
  int i, n;
  
  r = ring_inner(r);

  /* print out some info */
  n=1;
  do {
    fprintf(stdout,"Ring %d: %d points\n",n,r->size);
    r2 = r;
    do {
      fprintf(stdout, "  %d <0x%p>  ", r2->index, r2);
      for (i=0; i<r2->x_dim; i++)
	fprintf(stdout," %g",r2->x[i]);
      fprintf(stdout,"\n");
      r2 = r2->right;
    } while (r2 != r);
    r = r->out;
    n++;
  } while (r != NULL);
  
}


void
ring_plot(ring r, double *params, int color, int symbol)
{
  ring r2;

  /* draw only one ring of the whole structure */
  r2 = r;
  do {
    point_plot(r2->x,params,color,symbol);
    r2 = r2->right;
  } while (r2 != r);
}

void
ring_save(ring r, memory m , double *params, int color, int symbol,
	  int ring_type, int skip)
{
  int n_pts, c3[3], nrings=0, skipped = 0;
  ring r2;

  /* set up color stuff */
  c3[0] = -1; /* force use of pick color */
  c3[1] = color;
  c3[2] = symbol;
  
  /* find the inner ring */
  r = ring_inner(r);
  n_pts = r->size;

  /* count the number of rings */
  r2 = r;
  while (r2 != NULL)
    {
      nrings++;
      r2 = r2->out;
    }

  /* save each ring as a trajectory */
  if (0 == memory_start_new_flow(m, (nrings+4)/3, 0, 0, n_pts, 0, 1))
    {
      /* go through each ring */
      while (r != NULL)
	{
	  if (skipped-- == 0)
	    {
	      /* go through each element in ring */
	      r2 = r;
	      memory_add_point(m, NULL, params, c3, NULL, NULL, NULL, &ring_type);
	      do {
		/* add point */
		memory_add_point(m, r2->x, NULL, NULL, NULL, NULL, NULL, NULL);
		r2 = r2->right;
	      } while (r != r2);
	      memory_next_traj(m);
	      skipped = skip;
	    }
	  r2 = r;
	  r = r->out;
	}
      if (skipped != skip)
	{
	  /* go through each element in ring */
	  r = r2;
	  memory_add_point(m, NULL, params, c3, NULL, NULL, NULL, &ring_type);
	  do {
	    /* add point */
	    memory_add_point(m, r2->x, NULL, NULL, NULL, NULL, NULL, NULL);
	    r2 = r2->right;
	  } while (r != r2);
	}
    }
}



ring
ring_inner(ring r)
{
  while (r->index != 0) r=r->right;
  while(r->in) r = r->in;
  return r;
}


